@@ -1,11 +1,16 @@
+import argparse
+import logging
 import sys
 import json
+import typing
 import pathlib
 import itertools
 import collections
+import collections.abc
 import urllib.parse
 
 import attr
+import csvw
 from csvw.metadata import TableGroup, Table, Column, Link, Schema, is_url
 from csvw import datatypes
 from csvw.dsv import iterrows
@@ -21,40 +26,12 @@ from pycldf import orm
 
 __all__ = [
     'Dataset', 'Generic', 'Wordlist', 'ParallelText', 'Dictionary', 'StructureDataset',
-    'iter_datasets']
+    'iter_datasets', 'sniff']
 
 MD_SUFFIX = '-metadata.json'
 ORM_CLASSES = {cls.component_name(): cls for cls in orm.Object.__subclasses__()}
 
 
-def sniff(p):
-    with p.open('rb') as fp:
-        c = fp.read(10)
-        try:
-            c = c.decode('utf8').strip()
-        except UnicodeDecodeError:
-            return False
-        if not c.startswith('{'):
-            return False
-    try:
-        d = jsonlib.load(p)
-    except json.decoder.JSONDecodeError:
-        return False
-    return d.get('dc:conformsTo', '').startswith(TERMS_URL)
-
-
-def iter_datasets(d):
-    """
-    Discover CLDF datasets - by identifying metadata files - in a directory.
-
-    :param d: directory
-    :return: generator of `Dataset` instances.
-    """
-    for p in walk(d, mode='files'):
-        if sniff(p):
-            yield Dataset.from_metadata(p)
-
-
 @attr.s
 class Module(object):
     """
@@ -142,26 +119,303 @@ class Dataset(object):
     API to access a CLDF dataset.
     """
 
-    def __init__(self, tablegroup):
+    def __init__(self, tablegroup: csvw.TableGroup):
+        """
+        A :class:`~pycldf.dataset.Dataset` is initialized passing a TableGroup. For convenience \
+        methods to get such a TableGroup, see the factory methods
+
+        - :meth:`~pycldf.dataset.Dataset.in_dir`
+        - :meth:`~pycldf.dataset.Dataset.from_metadata`
+        - :meth:`~pycldf.dataset.Dataset.from_data`
+        """
         self.tablegroup = tablegroup
         self.auto_constraints()
         self.sources = Sources.from_file(self.bibpath)
         self._objects = collections.defaultdict(collections.OrderedDict)
+        self._objects_by_pk = collections.defaultdict(collections.OrderedDict)
+
+    #
+    # Factory methods to create `Dataset` instances.
+    #
+    @classmethod
+    def in_dir(cls, d: typing.Union[str, pathlib.Path], empty_tables=False):
+        """
+        Create a :class:`~pycldf.dataset.Dataset` in a (possibly empty or even non-existing) \
+        directory.
+
+        The dataset will be initialized with the default metadata for the respective module.
+
+        :rtype: :class:`~pycldf.dataset.Dataset`
+        """
+        fname = pathlib.Path(d)
+        if not fname.exists():
+            fname.mkdir()
+        assert fname.is_dir()
+        res = cls.from_metadata(fname)
+        if empty_tables:
+            del res.tables[:]
+        return res
+
+    @classmethod
+    def from_metadata(cls, fname: typing.Union[str, pathlib.Path]):
+        """
+        Initialize a :class:`~pycldf.dataset.Dataset` with the metadata found at `fname`.
+
+        :param fname: A URL (`str`) or a local path (`str` or `pathlib.Path`). \
+        If `fname` points to a directory, the default metadata for the respective module will be \
+        read.
+        :rtype: :class:`~pycldf.dataset.Dataset`
+        """
+        if is_url(fname):
+            tablegroup = TableGroup.from_url(fname)
+        else:
+            fname = pathlib.Path(fname)
+            if fname.is_dir():
+                name = '{0}{1}'.format(cls.__name__, MD_SUFFIX)
+                tablegroup = TableGroup.from_file(pkg_path('modules', name))
+                # adapt the path of the metadata file such that paths to tables are resolved
+                # correctly:
+                tablegroup._fname = fname.joinpath(name)
+            else:
+                tablegroup = TableGroup.from_file(fname)
+
+        comps = collections.Counter()
+        for table in tablegroup.tables:
+            try:
+                dt = Dataset.get_tabletype(table)
+                if dt:
+                    comps.update([dt])
+            except ValueError:
+                pass
+        if comps and comps.most_common(1)[0][1] > 1:
+            raise ValueError('{0}: duplicate components!'.format(fname))
+
+        for mod in get_modules():
+            if mod.match(tablegroup):
+                return mod.cls(tablegroup)
+        return cls(tablegroup)
+
+    @classmethod
+    def from_data(cls, fname):
+        """
+        Initialize a :class:`~pycldf.dataset.Dataset` from a single CLDF data file.
+
+        See https://github.com/cldf/cldf#metadata-free-conformance
+
+        :rtype: :class:`~pycldf.dataset.Dataset`
+        """
+        fname = pathlib.Path(fname)
+        colnames = next(iterrows(fname), [])
+        if not colnames:
+            raise ValueError('empty data file!')
+        if cls is Dataset:
+            try:
+                cls = next(mod.cls for mod in get_modules() if mod.match(fname))
+            except StopIteration:
+                raise ValueError('{0} does not match a CLDF module spec'.format(fname))
+            assert issubclass(cls, Dataset) and cls is not Dataset
+
+        res = cls.from_metadata(fname.parent)
+        required_cols = {
+            c.name for c in res[res.primary_table].tableSchema.columns
+            if c.required}
+        if not required_cols.issubset(colnames):
+            raise ValueError('missing columns: %r' % sorted(required_cols.difference(colnames)))
+        return res
+
+    #
+    # Accessing dataset metadata
+    #
+    @property
+    def directory(self) -> typing.Union[str, pathlib.Path]:
+        """
+        :return: The location of the metadata file. Either a local directory as `pathlib.Path` or \
+        a URL as `str`.
+        """
+        return self.tablegroup._fname.parent if self.tablegroup._fname else self.tablegroup.base
 
     @property
-    def metadata_dict(self):
+    def module(self) -> str:
+        """
+        :return: The name of the CLDF module of the dataset.
+        """
+        return self.properties['dc:conformsTo'].split('#')[1]
+
+    @property
+    def version(self):
+        return self.properties['dc:conformsTo'].split('/')[3]
+
+    def __repr__(self):
+        return '<cldf:%s:%s at %s>' % (self.version, self.module, self.directory)
+
+    @property
+    def metadata_dict(self) -> dict:
         return self.tablegroup.asdict(omit_defaults=False)
 
     @property
-    def properties(self):
+    def properties(self) -> dict:
+        """
+        :return: Common properties of the CSVW TableGroup of the dataset.
+        """
         return self.tablegroup.common_props
 
     @property
-    def tables(self):
+    def bibpath(self) -> typing.Union[str, pathlib.Path]:
+        """
+        :return: Location of the sources BibTeX file. Either a URL (`str`) or a local path \
+        (`pathlib.Path`).
+        """
+        # Specifying "dc:source": "" means lookup the default location.
+        if is_url(self.directory):
+            return urllib.parse.urljoin(
+                self.directory, self.properties.get('dc:source') or 'sources.bib')
+        return self.directory.joinpath(self.properties.get('dc:source') or 'sources.bib')
+
+    @property
+    def bibname(self) -> str:
+        """
+        :return: Filename of the sources BibTeX file.
+        """
+        if is_url(self.bibpath):
+            return pathlib.Path(urllib.parse.urlparse(self.bibpath).path).name
+        return self.bibpath.name
+
+    #
+    # Accessing schema objects (components, tables, columns, foreign keys)
+    #
+    @property
+    def tables(self) -> list:
+        """
+        :return: All tables defined in the dataset.
+        """
         return self.tablegroup.tables
 
     @property
-    def column_names(self):
+    def components(self) -> typing.Dict[str, csvw.Table]:
+        """
+        :return: Mapping of component name to table obejcts as defined in the dataset.
+        """
+        res = collections.OrderedDict()
+        for table in self.tables:
+            comp = None
+            try:
+                comp = self.get_tabletype(table)
+            except ValueError:
+                pass
+            if comp:
+                res[comp] = table
+        return res
+
+    @staticmethod
+    def get_tabletype(table) -> typing.Union[str, None]:
+        if table.common_props.get('dc:conformsTo', '') is None:
+            return None
+        if '#' in table.common_props.get('dc:conformsTo', ''):
+            res = table.common_props['dc:conformsTo'].split('#')[1]
+            if res in TERMS:
+                return res
+        raise ValueError("Type {:} of table {:} is not a valid term.".format(
+            table.common_props.get('dc:conformsTo'),
+            table.url))
+
+    @property
+    def primary_table(self) -> typing.Union[str, None]:
+        if self.tables:
+            try:
+                return self.get_tabletype(self.tables[0])
+            except ValueError:
+                return None
+
+    def __getitem__(self, item) -> typing.Union[csvw.Table, csvw.Column]:
+        """
+        Access to tables and columns.
+
+        If a pair (table-spec, column-spec) is passed as `item`, a Column will be
+        returned, otherwise `item` is assumed to be a table-spec.
+
+        A table-spec may be
+
+        - a CLDF ontology URI matching the dc:conformsTo property of a table
+        - the local name of a CLDF ontology URI, where the complete URI matches the \
+          the dc:conformsTo property of a table
+        - a filename matching the `url` property of a table
+
+        A column-spec may be
+
+        - a CLDF ontology URI matching the propertyUrl of a column
+        - the local name of a CLDF ontology URI, where the complete URI matches the \
+          propertyUrl of a column
+        - the name of a column
+
+        :raises KeyError: If no matching table or column is found.
+        """
+        if isinstance(item, tuple):
+            table, column = item
+        else:
+            table, column = item, None
+
+        if isinstance(table, Link):
+            table = table.string
+
+        if not isinstance(table, Table):
+            uri = term_uri(table, terms=TERMS.by_uri)
+            for t in self.tables:
+                if (uri and t.common_props.get('dc:conformsTo') == uri) \
+                        or t.url.string == table:
+                    break
+            else:
+                raise KeyError(table)
+        else:
+            t = table
+
+        if not column:
+            return t
+
+        uri = term_uri(column, terms=TERMS.by_uri)
+        for c in t.tableSchema.columns:
+            if (c.propertyUrl and c.propertyUrl.uri == uri) or c.header == column:
+                return c
+
+        raise KeyError(column)
+
+    def __contains__(self, item) -> bool:
+        """
+        Check whether a dataset specifies a table or column.
+
+        :param item: See  :meth:`~pycldf.dataset.Dataset.__getitem__`
+        """
+        return bool(self.get(item))
+
+    def get(self, item, default=None) -> typing.Union[csvw.Table, csvw.Column, None]:
+        """
+        Acts like `dict.get`.
+
+        :param item: See  :meth:`~pycldf.dataset.Dataset.__getitem__`
+        """
+        try:
+            return self[item]
+        except KeyError:
+            return default
+
+    def get_foreign_key_reference(self, table, column) \
+            -> typing.Union[typing.Tuple[csvw.Table, csvw.Column], None]:
+        """
+        Retrieve the reference of a foreign key constraint for the specified column.
+
+        :param table: Source table, specified by filename, component name or as `Table` instance.
+        :param column: Source column, specified by column name, CLDF term or as `Column` instance.
+        :return: A pair (`Table`, `Column`) specifying the reference column - or `None`.
+        """
+        table = self[table]
+        column = self[table, column]
+
+        for fk in table.tableSchema.foreignKeys:
+            if len(fk.columnReference) == 1 and fk.columnReference[0] == column.name:
+                return self[fk.reference.resource], \
+                    self[fk.reference.resource, fk.reference.columnReference[0]]
+
+    @property
+    def column_names(self) -> argparse.Namespace:
         """
         In-direction layer, mapping ontology terms to local column names (or `None`).
 
@@ -178,12 +432,15 @@ class Dataset(object):
         return get_column_names(self)
 
     @lazyproperty
-    def readonly_column_names(self):
+    def readonly_column_names(self) -> argparse.Namespace:
         """
         :return: `argparse.Namespace` with component names as attributes.
         """
         return get_column_names(self, use_component_names=True, with_multiplicity=True)
 
+    #
+    # Editing dataset metadata or schema
+    #
     def add_provenance(self, **kw):
         """
         Add metadata about the dataset's provenance.
@@ -213,18 +470,15 @@ class Dataset(object):
                 v = old
             self.tablegroup.common_props[k] = v
 
-    def add_sources(self, *sources, **kw):
-        self.sources.add(*sources, **kw)
-
-    def add_table(self, url, *cols, **kw):
+    def add_table(self, url: str, *cols, **kw) -> csvw.Table:
         """
         Add a table description to the Dataset.
 
         :param url: The url property of the table.
-        :param cols: Column specifications.
+        :param cols: Column specifications; anything accepted by :func:`pycldf.dataset.make_column`.
         :param kw: Recognized keywords:
             - `primaryKey`: specify the column(s) constituting the primary key of the table.
-        :return: `csvw.metadata.Table` instance.
+        :return: The new table.
         """
         t = self.add_component({"url": url, "tableSchema": {"columns": []}}, *cols)
         if 'primaryKey' in kw:
@@ -233,6 +487,9 @@ class Dataset(object):
         return t
 
     def remove_table(self, table):
+        """
+        Removes the table specified by `table` from the dataset.
+        """
         table = self[table]
 
         # First remove foreign keys:
@@ -243,11 +500,14 @@ class Dataset(object):
         # Now remove the table:
         self.tablegroup.tables = [t for t in self.tablegroup.tables if t.url != table.url]
 
-    def add_component(self, component, *cols, **kw):
+    def add_component(self, component: typing.Union[str, dict], *cols, **kw) -> csvw.Table:
         """
         Add a CLDF component to a dataset.
 
         .. seealso:: https://github.com/cldf/cldf/blob/master/README.md#cldf-components
+
+        :param component: A component specified by name or as `dict` representing the JSON \
+        description of the component.
         """
         if isinstance(component, str):
             component = jsonlib.load(pkg_path('components', '{0}{1}'.format(component, MD_SUFFIX)))
@@ -281,7 +541,10 @@ class Dataset(object):
         self.auto_constraints(component)
         return component
 
-    def add_columns(self, table, *cols):
+    def add_columns(self, table, *cols) -> None:
+        """
+        Add columns specified by `cols` to the table specified by `table`.
+        """
         table = self[table]
         for col in cols:
             existing = [c.name for c in table.tableSchema.columns]
@@ -292,7 +555,7 @@ class Dataset(object):
                 raise ValueError('Duplicate column name: {0}'.format(col.name))
             if col.propertyUrl and col.propertyUrl.uri in existing:
                 raise ValueError('Duplicate column property: {0}'.format(col.propertyUrl.uri))
-            table.tableSchema.columns.append(make_column(col))
+            table.tableSchema.columns.append(col)
         self.auto_constraints()
 
     def remove_columns(self, table, *cols):
@@ -370,7 +633,7 @@ class Dataset(object):
             # `xxxReference` column properties?
             return
 
-        # auto-add foreign keys targetting the new component:
+        # auto-add foreign keys targeting the new component:
         for table in self.tables:
             self._auto_foreign_keys(table, component=component, table_type=table_type)
 
@@ -397,21 +660,160 @@ class Dataset(object):
                 table.add_foreign_key(
                     col.name, ref.url.string, idcol.name if idcol is not None else 'ID')
 
-    @property
-    def bibpath(self):
-        # Specifying "dc:source": "" means lookup the default location.
-        if is_url(self.directory):
-            return urllib.parse.urljoin(
-                self.directory, self.properties.get('dc:source') or 'sources.bib')
-        return self.directory.joinpath(self.properties.get('dc:source') or 'sources.bib')
+    #
+    # Add data
+    #
+    def add_sources(self, *sources, **kw):
+        """
+        Add sources to the dataset.
 
-    @property
-    def bibname(self):
-        if is_url(self.bibpath):
-            return pathlib.Path(urllib.parse.urlparse(self.bibpath).path).name
-        return self.bibpath.name
+        :param sources: Anything accepted by :meth:`pycldf.sources.Sources.add`.
+        """
+        self.sources.add(*sources, **kw)
 
-    def validate(self, log=None, validators=None, ontology_path=None):
+    #
+    # Methods to read data
+    #
+    def iter_rows(self, table, *cols) -> typing.Iterable[dict]:
+        """
+        Iterate rows in a table, resolving CLDF property names to local column names.
+
+        :param table: Table name.
+        :param cols: List of CLDF property terms which must be resolved in resulting `dict` s. \
+        I.e. the row dicts will be augmented with copies of the values keyed with CLDF property \
+        terms.
+        """
+        cmap = {self[table, col].name: col for col in cols}
+        for item in self[table]:
+            for k, v in cmap.items():
+                # Add CLDF properties as aliases for the corresponding values:
+                item[v] = item[k]
+            yield item
+
+    def get_row(self, table, id_) -> dict:
+        """
+        Retrieve a row specified by table and CLDF id.
+
+        :raises ValueError: If no matching row is found.
+        """
+        id_col = self[table, TERMS['id']]
+        for row in self[table]:
+            if row[id_col.name] == id_:
+                return row
+        raise ValueError(id_)  # pragma: no cover
+
+    def get_row_url(self, table, row) -> typing.Union[str, None]:
+        """
+        Get a URL associated with a row. Tables can specify associated row URLs by
+
+        - listing **one** column with datatype `anyURI` or
+        - specfying a `valueUrl` property for their ID column.
+
+        For rows representing objects in web applications, this may be the objects URL. For
+        rows representing media files, it may be a URL locating the file on a media server.
+
+        :param table: Table specified in a way that `__getitem__` understands.
+        :param row: A row specified by ID or as `dict` as returned when iterating over a table.
+        :return: a `str` representing a URL or `None`.
+        """
+        row = row if isinstance(row, dict) else self.get_row(table, row)
+        id_col = None
+        for col in self[table].tableSchema.columns:
+            if col.datatype and col.datatype.base == datatypes.anyURI.__name__:
+                # If one of the columns in the table explicitly spacifies anyURI as datatype, we
+                # return the value of this column.
+                return row[col.name]
+            if str(col.propertyUrl) == 'http://cldf.clld.org/v1.0/terms.rdf#id':
+                # Otherwise we fall back to looking up the `valueUrl` property on the ID column.
+                id_col = col
+        assert id_col, 'no ID column found in table {}'.format(table)
+        if id_col.valueUrl:
+            return id_col.valueUrl.expand(**row)
+
+    def objects(self, table, cls=None) -> DictTuple:
+        """
+        Read data of a CLDF component as :class:`pycldf.orm.Object` instances.
+
+        :param table: table to read, specified as component name.
+        :param cls: :class:`pycldf.orm.Object` subclass to instantiate objects with.
+        :return:
+        """
+        cls = cls or ORM_CLASSES[table]
+
+        # ORM usage is read-only, so we can cache the objects.
+        if table not in self._objects:
+            for item in self[table]:
+                item = cls(self, item)
+                self._objects[table][item.id] = item
+                if item.pk:
+                    self._objects_by_pk[table][item.pk] = item
+
+        return DictTuple(self._objects[table].values())
+
+    def get_object(self, table, id_, cls=None, pk=False) -> orm.Object:
+        """
+        Get a row of a component as :class:`pycldf.orm.Object` instance.
+        """
+        if table not in self._objects:
+            self.objects(table, cls=cls)
+        return self._objects[table][id_] if not pk else self._objects_by_pk[table][id_]
+
+    #
+    # Methods for writing (meta)data to files:
+    #
+    def write_metadata(self, fname=None):
+        """
+        Write the CLDF metadata to a JSON file.
+
+        :fname: Path of a file to write to, or `None` to use the default name and write to \
+        :meth:`~pycldf.dataset.Dataset.directory`.
+        """
+        return self.tablegroup.to_file(fname or self.tablegroup._fname)
+
+    def write_sources(self):
+        """
+        Write the sources BibTeX file to :meth:`~pycldf.dataset.Dataset.bibpath`
+        """
+        return self.sources.write(self.bibpath)
+
+    def write(self, fname=None, **table_items: typing.Dict[str, typing.List[dict]]):
+        """
+        Write metadata, sources and data. Metadata will be written to `fname` (as interpreted in
+        :meth:`pycldf.dataset.Dataset.write_metadata`); data files will be written to the file
+        specified by `csvw.Table.url` of the corresponding table, interpreted as path relative
+        to :meth:`~pycldf.dataset.Dataset.directory`.
+
+        :param table_items: Mapping of table specifications to lists of row dicts.
+        """
+        if self.sources and not self.properties.get('dc:source'):
+            self.properties['dc:source'] = 'sources.bib'
+        self.write_sources()
+        for table_type, items in table_items.items():
+            table = self[table_type]
+            table.common_props['dc:extent'] = table.write(items)
+        self.write_metadata(fname)
+
+    #
+    # Reporting
+    #
+    def validate(
+            self,
+            log: logging.Logger = None,
+            validators: typing.List[typing.Tuple[str, str, callable]] = None,
+            ontology_path=None) -> bool:
+        """
+        Validate schema and data of a `Dataset`:
+
+        - Make sure the schema follows the CLDF specification and
+        - make sure the data is consistent with the schema.
+
+        :param log: a `logging.Logger` to write ERRORs and WARNINGs to. If `None`, an exception \
+        will be raised at the first problem.
+        :param validators: Custom validation rules, i.e. triples \
+        (tablespec, columnspec, attrs validator)
+        :raises ValueError: if a validation error is encountered (and `log` is `None`).
+        :return: Flag signaling whether schema and data are valid.
+        """
         terms = Terms(ontology_path) or TERMS
         validators = validators or []
         validators.extend(VALIDATORS)
@@ -521,200 +923,12 @@ class Dataset(object):
 
         return success
 
-    @property
-    def directory(self):
-        return self.tablegroup._fname.parent if self.tablegroup._fname else self.tablegroup.base
-
-    @property
-    def module(self):
-        return self.properties['dc:conformsTo'].split('#')[1]
-
-    @property
-    def version(self):
-        return self.properties['dc:conformsTo'].split('/')[3]
-
-    @classmethod
-    def in_dir(cls, d, empty_tables=False):
-        fname = pathlib.Path(d)
-        if not fname.exists():
-            fname.mkdir()
-        assert fname.is_dir()
-        res = cls.from_metadata(fname)
-        if empty_tables:
-            del res.tables[:]
-        return res
-
-    #
-    # Factory methods to create `Dataset` instances.
-    #
-    @classmethod
-    def from_metadata(cls, fname):
-        if is_url(fname):
-            tablegroup = TableGroup.from_url(fname)
-        else:
-            fname = pathlib.Path(fname)
-            if fname.is_dir():
-                name = '{0}{1}'.format(cls.__name__, MD_SUFFIX)
-                tablegroup = TableGroup.from_file(pkg_path('modules', name))
-                # adapt the path of the metadata file such that paths to tables are resolved
-                # correctly:
-                tablegroup._fname = fname.joinpath(name)
-            else:
-                tablegroup = TableGroup.from_file(fname)
-
-        comps = collections.Counter()
-        for table in tablegroup.tables:
-            try:
-                dt = Dataset.get_tabletype(table)
-                if dt:
-                    comps.update([dt])
-            except ValueError:
-                pass
-        if comps and comps.most_common(1)[0][1] > 1:
-            raise ValueError('{0}: duplicate components!'.format(fname))
-
-        for mod in get_modules():
-            if mod.match(tablegroup):
-                return mod.cls(tablegroup)
-        return cls(tablegroup)
-
-    @classmethod
-    def from_data(cls, fname):
-        fname = pathlib.Path(fname)
-        colnames = next(iterrows(fname), [])
-        if not colnames:
-            raise ValueError('empty data file!')
-        if cls is Dataset:
-            try:
-                cls = next(mod.cls for mod in get_modules() if mod.match(fname))
-            except StopIteration:
-                raise ValueError('{0} does not match a CLDF module spec'.format(fname))
-            assert issubclass(cls, Dataset) and cls is not Dataset
-
-        res = cls.from_metadata(fname.parent)
-        required_cols = {
-            c.name for c in res[res.primary_table].tableSchema.columns
-            if c.required}
-        if not required_cols.issubset(colnames):
-            raise ValueError('missing columns: %r' % sorted(required_cols.difference(colnames)))
-        return res
-
-    def __repr__(self):
-        return '<cldf:%s:%s at %s>' % (self.version, self.module, self.directory)
-
-    def __getitem__(self, item):
+    def stats(self, exact=False) -> typing.List[typing.Tuple[str, str, int]]:
         """
-        Access to tables and columns.
+        Compute summary statistics for the dataset.
 
-        If a pair (table-spec, column-spec) is passed as `item`, a Column will be
-        returned, otherwise `item` is assumed to be a table-spec.
-
-        A table-spec may be
-        - a CLDF ontology URI matching the dc:conformsTo property of a table
-        - the local name of a CLDF ontology URI, where the complete URI matches the
-          the dc:conformsTo property of a table
-        - a filename matching the `url` property of a table
-
-        A column-spec may be
-        - a CLDF ontology URI matching the propertyUrl of a column
-        - the local name of a CLDF ontology URI, where the complete URI matches the
-          propertyUrl of a column
-        - the name of a column
+        :return: List of triples (table, type, rowcount).
         """
-        if isinstance(item, tuple):
-            table, column = item
-        else:
-            table, column = item, None
-
-        if not isinstance(table, Table):
-            uri = term_uri(table, terms=TERMS.by_uri)
-            for t in self.tables:
-                if (uri and t.common_props.get('dc:conformsTo') == uri) \
-                        or t.url.string == table:
-                    break
-            else:
-                raise KeyError(table)
-        else:
-            t = table
-
-        if not column:
-            return t
-
-        uri = term_uri(column, terms=TERMS.by_uri)
-        for c in t.tableSchema.columns:
-            if (c.propertyUrl and c.propertyUrl.uri == uri) or c.header == column:
-                return c
-
-        raise KeyError(column)
-
-    def __contains__(self, item):
-        try:
-            self.__getitem__(item)
-            return True
-        except KeyError:
-            return False
-
-    def get(self, item, default=None):
-        try:
-            return self[item]
-        except KeyError:
-            return default
-
-    def get_row(self, table, id_):
-        id_col = self[table, TERMS['id']]
-        for row in self[table]:
-            if row[id_col.name] == id_:
-                return row
-        raise ValueError(id_)  # pragma: no cover
-
-    def get_row_url(self, table, row):
-        """
-        Get a URL associated with a row. Tables can specify associated row URLs by
-        - listing **one** column with datatype `anyURI` or
-        - specfying a `valueUrl` property for their ID column.
-
-        For rows representing objects in web applications, this may be the objects URL. For
-        rows representing media files, it may be a URL locating the file on a media server.
-
-        :param table: Table specified in a way that `__getitem__` understands.
-        :param row: A row specified by ID or as `dict` as returned when iterating over a table.
-        :return: a `str` representing a URL or `None`.
-        """
-        row = row if isinstance(row, dict) else self.get_row(table, row)
-        id_col = None
-        for col in self[table].tableSchema.columns:
-            if col.datatype and col.datatype.base == datatypes.anyURI.__name__:
-                # If one of the columns in the table explicitly spacifies anyURI as datatype, we
-                # return the value of this column.
-                return row[col.name]
-            if str(col.propertyUrl) == 'http://cldf.clld.org/v1.0/terms.rdf#id':
-                # Otherwise we fall back to looking up the `valueUrl` property on the ID column.
-                id_col = col
-        assert id_col, 'no ID column found in table {}'.format(table)
-        if id_col.valueUrl:
-            return id_col.valueUrl.expand(**row)
-
-    @staticmethod
-    def get_tabletype(table):
-        if table.common_props.get('dc:conformsTo', '') is None:
-            return None
-        if '#' in table.common_props.get('dc:conformsTo', ''):
-            res = table.common_props['dc:conformsTo'].split('#')[1]
-            if res in TERMS:
-                return res
-        raise ValueError("Type {:} of table {:} is not a valid term.".format(
-            table.common_props.get('dc:conformsTo'),
-            table.url))
-
-    @property
-    def primary_table(self):
-        if self.tables:
-            try:
-                return self.get_tabletype(self.tables[0])
-            except ValueError:
-                return None
-
-    def stats(self, exact=False):
         res = []
         for table in self.tables:
             dctype = table.common_props.get('dc:conformsTo')
@@ -729,59 +943,6 @@ class Dataset(object):
             res.append((self.bibname, 'Sources', len(self.sources)))
         return res
 
-    def write_metadata(self, fname=None):
-        return self.tablegroup.to_file(fname or self.tablegroup._fname)
-
-    def write_sources(self):
-        return self.sources.write(self.bibpath)
-
-    def write(self, fname=None, **table_items):
-        if self.sources and not self.properties.get('dc:source'):
-            self.properties['dc:source'] = 'sources.bib'
-        self.write_sources()
-        for table_type, items in table_items.items():
-            table = self[table_type]
-            table.common_props['dc:extent'] = table.write(items)
-        self.write_metadata(fname)
-
-    def iter_rows(self, table, *cols):
-        """
-        Iterate rows in a table, resolving CLDF property names to local column names.
-
-        :param table: Table name.
-        :param cols: List of CLDF property terms which must be resolved in resulting `dict`s.
-        :return: Generator
-        """
-        cmap = {self[table, col].name: col for col in cols}
-        for item in self[table]:
-            for k, v in cmap.items():
-                # Add CLDF properties as aliases for the corresponding values:
-                item[v] = item[k]
-            yield item
-
-    def objects(self, table, cls=None):
-        """
-        Read data of a CLDF component as `pycldf.orm.Object` instances.
-
-        :param table: table to read, specified as component name.
-        :param cls: `orm.Object` subclass to instantiate objects with.
-        :return:
-        """
-        cls = cls or ORM_CLASSES[table]
-
-        # ORM usage is read-only, so we can cache the objects.
-        if table not in self._objects:
-            for item in self[table]:
-                item = cls(self, item)
-                self._objects[table][item.id] = item
-
-        return DictTuple(self._objects[table].values())
-
-    def get_object(self, table, id_, cls=None):
-        if table not in self._objects:
-            self.objects(table, cls=cls)
-        return self._objects[table][id_]
-
 
 class Generic(Dataset):
     @property
@@ -845,3 +1006,37 @@ class StructureDataset(Dataset):
     @property
     def primary_table(self):
         return 'ValueTable'
+
+
+def sniff(p: pathlib.Path) -> bool:
+    """
+    Determine whether a file contains CLDF metadata.
+
+    :param p: `pathlib.Path` object for an existing file.
+    :return: `True` if the file contains CLDF metadata, `False` otherwise.
+    """
+    with p.open('rb') as fp:
+        c = fp.read(10)
+        try:
+            c = c.decode('utf8').strip()
+        except UnicodeDecodeError:
+            return False
+        if not c.startswith('{'):
+            return False
+    try:
+        d = jsonlib.load(p)
+    except json.decoder.JSONDecodeError:
+        return False
+    return d.get('dc:conformsTo', '').startswith(TERMS_URL)
+
+
+def iter_datasets(d: pathlib.Path) -> typing.Iterable[Dataset]:
+    """
+    Discover CLDF datasets - by identifying metadata files - in a directory.
+
+    :param d: directory
+    :return: generator of `Dataset` instances.
+    """
+    for p in walk(d, mode='files'):
+        if sniff(p):
+            yield Dataset.from_metadata(p)
