@@ -2,49 +2,76 @@
 Object oriented (read-only) access to CLDF data
 
 To read ORM objects from a `pycldf.Dataset`, use two methods
-- `pycldf.Dataset.objects`
-- `pycldf.Dataset.get_object`
+
+* `pycldf.Dataset.objects`
+* `pycldf.Dataset.get_object`
 
 Both will return default implementations of the objects, i.e. instances of the corresponding
 class defined in this module. To customize these objects,
-- subclass the default and specify
-  the appropriate component (i.e. the table of the CLDF dataset which holds rows to be transformed
-  to this type):
 
-  ```python
-  from pycldf.orm import Language
+1. subclass the default and specify \
+   the appropriate component (i.e. the table of the CLDF dataset which holds rows to be transformed\
+   to this type):
 
-  class Variety(Language):
-      __component__ = 'LanguageTable'
+   .. code-block:: python
 
-      def custom_method(self):
-          pass
-  ```
-- pass the class into the `objects` or `get_object` method.
+      from pycldf.orm import Language
 
-This functionality comes with the typical "more convenient API vs. less performance and bigger
-memory footprint" trade-off. If you are running into problems with this, you might want to load
-your data into a SQLite db using the `pycldf.db` module, and access via SQL.
+      class Variety(Language):
+          __component__ = 'LanguageTable'
+
+          def custom_method(self):
+              pass
+
+2. pass the class into the `objects` or `get_object` method.
+
+Limitations:
+------------
+* We only support foreign key constraints for CLDF reference properties targeting either a \
+  component's CLDF id or its primary key. This is because CSVW does not support unique constraints \
+  other than the one implied by the primary key declaration.
+* This functionality comes with the typical "more convenient API vs. less performance and bigger \
+  memory footprint" trade-off. If you are running into problems with this, you might want to load \
+  your data into a SQLite db using the `pycldf.db` module, and access via SQL. \
+  Some numbers (to be interpreted relative to each other): \
+  Reading ~400,000 rows from a ValueTable of a StructureDataset takes
+
+  * ~2secs with csvcut, i.e. only making sure it's valid CSV
+  * ~15secs iterating over pycldf.Dataset['ValueTable']
+  * ~35secs iterating over pycldf.Dataset.objects('ValueTable')
 """
 import argparse
 import collections
+import typing
 
+import csvw.metadata
 from tabulate import tabulate
 from clldutils.misc import lazyproperty
 
-from pycldf.terms import TERMS
+from pycldf.terms import TERMS, term_uri
 from pycldf.util import DictTuple
+from pycldf.sources import Reference
 
 
 class Object:
     """
     Represents a row of a CLDF component table.
+
+    Subclasses of `Object` are instantiated when calling `Dataset.objects` or `Dataset.get_object`.
+
+    :ivar dataset: Reference to the `Dataset` instance, this object was loaded from.
+    :ivar data: An `OrderedDict` with a copy of the row the object was instantiated with.
+    :ivar cldf: A `dict` with CLDF-specified properties of the row, keyed with CLDF terms.
+    :ivar id: The value of the CLDF id property of the row.
+    :ivar name: The value of the CLDF name property of the row.
+    :ivar description: The value of the CLDF description property of the row.
+    :ivar pk: The value of the column specified as primary key for the table. (May differ from id)
     """
     # If a subclass name can not be used to derive the CLDF component name, the component can be
     # specified here:
     __component__ = None
 
-    def __init__(self, dataset, row):
+    def __init__(self, dataset, row: dict):
         # Get a mapping of column names to pairs (CLDF property name, list-valued) for columns
         # present in the component specified by class name.
         cldf_cols = {
@@ -63,15 +90,26 @@ class Object:
         self.cldf = argparse.Namespace(**self.cldf)
         self.dataset = dataset
         self.id = self.cldf.id
+        self.pk = None
+        t = dataset[self.component_name()]
+        if t.tableSchema.primaryKey and len(t.tableSchema.primaryKey) == 1:
+            self.pk = self.data[dataset[self.component_name()].tableSchema.primaryKey[0]]
         self.name = getattr(self.cldf, 'name', None)
         self.description = getattr(self.cldf, 'name', None)
 
+    def __repr__(self):
+        return '<{}.{} id="{}">'.format(self.__class__.__module__, self.__class__.__name__, self.id)
+
     @classmethod
-    def component_name(cls):
+    def component_name(cls) -> str:
         return cls.__component__ or (cls.__name__ + 'Table')
 
     @property
-    def component(self):
+    def component(self) -> str:
+        """
+        Name of the CLDF component the object belongs to. Can be used to lookup the corresponding \
+        table via `obj.dataset[obj.component_name()]`.
+        """
         return self.__class__.component_name()
 
     @property
@@ -97,23 +135,32 @@ class Object:
         if getattr(col, attr, None):
             return getattr(col, attr).expand(**variables)
 
-    def aboutUrl(self, col='id'):
+    def aboutUrl(self, col='id') -> typing.Union[str, None]:
+        """
+        The table's `aboutUrl` property, expanded with the object's row as context.
+        """
         return self._expand_uritemplate('aboutUrl', col)
 
     def valueUrl(self, col='id'):
+        """
+        The table's `valueUrl` property, expanded with the object's row as context.
+        """
         return self._expand_uritemplate('valueUrl', col)
 
     def propertyUrl(self, col='id'):
+        """
+        The table's `propertyUrl` property, expanded with the object's row as context.
+        """
         return self._expand_uritemplate('propertyUrl', col)
 
     @lazyproperty
-    def references(self):
+    def references(self) -> typing.Tuple[Reference]:
         """
+        `pycldf.Reference` instances associated with the object.
+
         >>> obj.references[0].source['title']
         >>> obj.references[0].fields.title
         >>> obj.references[0].description  # The "context", typically cited pages
-
-        :return: `list` of `pycldf.sources.Reference` objects.
         """
         return DictTuple(
             self.dataset.sources.expand_refs(getattr(self.cldf, 'source', []) or []),
@@ -121,7 +168,7 @@ class Object:
             multi=True,
         )
 
-    def related(self, relation):
+    def related(self, relation: str):
         """
         The CLDF ontology specifies several "reference properties". This method returns the first
         related object specified by such a property.
@@ -134,9 +181,16 @@ class Object:
                 '{} is list-valued, use `all_related` to retrieve related objects'.format(relation))
         fk = getattr(self.cldf, relation, None)
         if fk:
-            return self.dataset.get_object(TERMS[relation].references, fk)
+            ref = self.dataset.get_foreign_key_reference(self.component_name(), relation)
+            if ref:
+                if str(ref[1].propertyUrl) == term_uri('id'):
+                    return self.dataset.get_object(TERMS[relation].references, fk)
+                if [ref[1].name] == self.dataset[TERMS[relation].references].tableSchema.primaryKey:
+                    return self.dataset.get_object(TERMS[relation].references, fk, pk=True)
+                raise NotImplementedError('pycldf does not support foreign key constraints '
+                                          'referencing columns other than CLDF id or primary key.')
 
-    def all_related(self, relation):
+    def all_related(self, relation: str) -> typing.Union[DictTuple, list]:
         """
         CLDF reference properties can be list-valued. This method returns all related objects for
         such a property.
@@ -149,7 +203,7 @@ class Object:
         return []
 
 
-class WithLanguageMixin:
+class _WithLanguageMixin:
     @property
     def language(self):
         return self.related('languageReference')
@@ -159,8 +213,8 @@ class WithLanguageMixin:
         return self.all_related('languageReference')
 
 
-class WithParameterMixin:
-    @property
+class _WithParameterMixin:
+    @lazyproperty
     def parameter(self):
         return self.related('parameterReference')
 
@@ -179,7 +233,7 @@ class Borrowing(Object):
         return self.related('sourceFormReference')
 
 
-class Code(Object, WithParameterMixin):
+class Code(Object, _WithParameterMixin):
     pass
 
 
@@ -197,13 +251,13 @@ class Cognate(Object):
         return self.related('cognatesetReference')
 
 
-class Entry(Object, WithLanguageMixin):
+class Entry(Object, _WithLanguageMixin):
     @property
     def senses(self):
         return DictTuple(v for v in self.dataset.objects('SenseTable') if self in v.entries)
 
 
-class Example(Object, WithLanguageMixin):
+class Example(Object, _WithLanguageMixin):
     @property
     def metaLanguage(self):
         return self.related('metaLanguageReference')
@@ -217,7 +271,7 @@ class Example(Object, WithLanguageMixin):
         )
 
 
-class Form(Object, WithLanguageMixin, WithParameterMixin):
+class Form(Object, _WithLanguageMixin, _WithParameterMixin):
     pass
 
 
@@ -271,6 +325,13 @@ class Language(Object):
 
 
 class Parameter(Object):
+    @lazyproperty
+    def datatype(self):
+        if 'datatype' in self.data \
+                and self.dataset['ParameterTable', 'datatype'].datatype.base == 'json':
+            if self.data['datatype']:
+                return csvw.metadata.Datatype.fromvalue(self.data['datatype'])
+
     @property
     def values(self):
         return DictTuple(v for v in self.dataset.objects('ValueTable') if self in v.parameters)
@@ -302,7 +363,13 @@ class Sense(Object):
         return self.all_related('entryReference')
 
 
-class Value(Object, WithLanguageMixin, WithParameterMixin):
+class Value(Object, _WithLanguageMixin, _WithParameterMixin):
+    @property
+    def typed_value(self):
+        if self.parameter.datatype:
+            return self.parameter.datatype.read(self.cldf.value)
+        return self.cldf.value
+
     @property
     def code(self):
         return self.related('codeReference')
